package com.framsticks.framclipse;

import org.antlr.runtime.CharStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.Token;

import com.framsticks.framclipse.parser.antlr.internal.InternalFramScriptLexer;

public class FramScriptAuxiliaryLexer extends InternalFramScriptLexer {

	private static final int FLG_MULTILINE = 0x1;
	private static final int FLG_TYPE_IS_NUMERIC = 0x2;	// if raised, the default value should be analysed with internal lexer
	private static final int FLG_TYPE_IS_STRING = 0x4;	// if raised, the max value can be omitted and should be 0
	private static final int CTX_NONE = 0;
	private static final int CTX_TYPE = 1;
	private static final int CTX_TYPE_ID = 2;
	private static final int CTX_TYPE_MIN = 3;
	private static final int CTX_TYPE_MAX = 4;
	private static final int CTX_TYPE_DEFAULT = 5;

	private int context = CTX_NONE;
	private int flags = 0;
	private Token previous = Token.INVALID_TOKEN;

	private void checkML() {
		setFlag(FLG_MULTILINE, next() == '~');
	}

	private boolean isSet(int flag) {
		return (flags & flag) != 0;
	}

	private void setFlag(int flag, boolean state) {
		if (state)
			flags |= flag;
		else
			flags &= ~flag;
	}

	private void setState(int tokenType) {
		state.token = null;
		state.channel = Token.DEFAULT_CHANNEL;
		state.tokenStartCharIndex = input.index();
		state.tokenStartCharPositionInLine = input.getCharPositionInLine();
		state.tokenStartLine = input.getLine();
		state.text = null;
		state.type = tokenType;
	}

	char next() {
		return next(1);
	}

	char next(int i) {
		return (char)input.LA(i);
	}

	void consumeUntilAny(String term) {
		term += (char)CharStream.EOF;
		while (term.indexOf(next()) == -1)
			matchAny();
	}

	@Override
	public Token nextToken() {
		switch (previous.getType()) {
		case RULE_HEADER_ID:
			setState(RULE_HEADER_VALUE);
			checkML();
			mHEADER_VALUE();
			return previous = emit();
		case RULE_TYPE_HEADER:
			checkML();
			context = CTX_TYPE;
			if (isSet(FLG_MULTILINE)) {
				setState(RULE_WS);	// any ignored token
				matchAny();	// consume '~'
				return previous = emit();
			}
			break;
		case RULE_WS:
			if (context == CTX_TYPE_MIN) {
				if (isSet(FLG_TYPE_IS_STRING)) {
					if (!Character.isDigit(next()) && next() != '-') {	// if string length is omitted
						if (isSet(FLG_MULTILINE))
							matchAny();	// ignore expected '~'
						setState(RULE_INT);
						state.text = "0";	// pretend the max value is 0
						return previous = emit();
					}
				}
			}
			if (previous.getText().indexOf('\n') != -1 ||
					previous.getText().indexOf('\r') != -1) {
				if (!isSet(FLG_MULTILINE)) {
					context = CTX_NONE;
				}
				// workaround for '#include'
				if (willMatch("#include")) {
					setState(RULE_PROP_INCLUDE);
					try {
						mRULE_PROP_INCLUDE();
					} catch (RecognitionException e) {
						reportError(e);
					}
					return previous = emit();
				}
				if (!isSet(FLG_MULTILINE)) {
					break;
				}
			}
			if (context == CTX_TYPE_MAX) {
				if (isSet(FLG_MULTILINE)) {
					context = CTX_TYPE_DEFAULT;
				} else {
					context = CTX_NONE;
				}
				if (isSet(FLG_TYPE_IS_NUMERIC)) {
					if (isSet(FLG_MULTILINE) && !Character.isDigit(next())
							&& next() != '-') {	// if default value is omitted
						previous = aux_ENUM_LITERAL();
						if (previous != null)
							return previous;
					}
					break;
				} else {
					setState(RULE_HEADER_VALUE);
					mRULE_TYPE_HEADER_VALUE();
					return previous = emit();
				}
			} else if (context == CTX_TYPE_DEFAULT) {
				previous = aux_ENUM_LITERAL();
				if (previous != null)
					return previous;
			}
			break;
		case RULE_ID:
			if (context == CTX_TYPE) {
				if ("df".indexOf(previous.getText().charAt(0)) == -1) {
					setFlag(FLG_TYPE_IS_NUMERIC, false);
					setFlag(FLG_TYPE_IS_STRING, previous.getText().charAt(0) == (int)'s');
				} else {
					setFlag(FLG_TYPE_IS_NUMERIC, true);
				}
				context = CTX_TYPE_ID;
				break;
			}
			context = CTX_NONE;
			break;
		case RULE_HEADER_VALUE:
		case RULE_ENUM_LITERAL:
			if (context == CTX_TYPE_DEFAULT) {
				previous = aux_ENUM_LITERAL();
				if (previous != null)
					return previous;
			}
			context = CTX_NONE;
			break;
		case RULE_DOUBLE:
		case RULE_INT:
			if (context == CTX_TYPE_DEFAULT) {
				break;	// avoid changing context
			} else if (context == CTX_TYPE_ID) {
				context = CTX_TYPE_MIN;
				break;
			} else if (context == CTX_TYPE_MIN) {
				context = CTX_TYPE_MAX;
				break;
			}
			context = CTX_NONE;
			break;
		default:
			if ("-".equals(previous.getText()) ||
					"+".equals(previous.getText())) {
				break;
			}
			context = CTX_NONE;
		case RULE_SL_COMMENT:
			break;
		}
		return previous = super.nextToken();
	}

	private boolean willMatch(String string) {
		for (int i = 0; i < string.length(); ++i) {
			if (next(i+1) != string.charAt(i))
				return false;
		}
		return true;
	}

	private Token aux_ENUM_LITERAL() {
		// CTX_TYPE_DEFAULT assumes raised FLG_MULTILINE
		if (next() == '~') {
			setState(RULE_WS);	// any ignored token
			matchAny();
			context = CTX_NONE;
			return emit();
		} else if (next() == '\\' && next(2) == '~') {
			matchAny();	// not assigned to any terminal
			setState(RULE_ENUM_LITERAL);
			matchAny();	// consume '~'
			mRULE_TYPE_HEADER_VALUE();
			return emit();
		}
		return null;
	}
	
	private void mRULE_TYPE_HEADER_VALUE() {
		if (isSet(FLG_MULTILINE)) {
			do {
				consumeUntilAny("\\~");
				if (next() == '\\' && next(2) != '~') {
					matchAny();
					continue;
				}
			} while (false);
		} else {
			consumeUntilAny("~\r\n");
		}
	}

	private void mHEADER_VALUE() {
		if (isSet(FLG_MULTILINE)) {
			matchAny();	// consume '~'
			do {
				consumeUntilAny("\\~");
				if (next() == '\\') {
					matchAny();
					if (next() == '~')
						matchAny();
					continue;
				}
			} while (false);
			matchAny();
		} else {
			consumeUntilAny("\r\n");
		}
	}
}
